﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using System.Threading;
using System.Diagnostics;

namespace Benday.SequenceGenerator.Business
{
    public class Sequence
    {
        private const int MAX_ATTEMPT_COUNT = 10;
        private const int RETRY_WAIT_TIME_IN_MILLISECONDS = 6000;
        private const string QUEUE_NAME_SEQUENCE_MANAGER = "sequencemanager";
        
        internal Sequence(string name)
        {
            if (String.IsNullOrEmpty(name) == true)
            {
                throw new ArgumentException("name is null or empty.", "name");
            }
            else if (name.Contains(":") == true)
            {
                throw new ArgumentException("Sequence name cannot contain ':'.", "name");
            }

            m_name = name;

            InitializeSequence();
        }

        private string m_name;
        public string Name
        {
            get { return m_name; }
        }

        private CloudQueueClient m_client;
        public CloudQueueClient Client
        {
            get
            {
                if (m_client == null)
                {
                    m_client = CloudStorageAccount.DevelopmentStorageAccount.CreateCloudQueueClient();
                }

                return m_client;
            }
        }

        private CloudQueue m_sequenceQueue;
        public CloudQueue SequenceQueue
        {
            get
            {
                if (m_sequenceQueue == null)
                {
                    m_sequenceQueue = Client.GetQueueReference(String.Format("sequence-{0}", Name.ToLower()));
                }

                return m_sequenceQueue;
            }
        }

        private CloudQueue m_managerQueue;
        public CloudQueue ManagerQueue
        {
            get
            {
                if (m_managerQueue == null)
                {
                    m_managerQueue = Client.GetQueueReference(QUEUE_NAME_SEQUENCE_MANAGER);
                    m_managerQueue.CreateIfNotExist();
                }

                return m_managerQueue;
            }
        }

        private void InitializeSequence()
        {
            bool queueExists = SequenceQueue.Exists();

            if (queueExists == false)
            {
                InitializeSequenceQueue(Name, Client, SequenceQueue);
            }
        }

        private void InitializeSequenceQueue(string sequenceName, CloudQueueClient client, CloudQueue queue)
        {
            Trace.TraceInformation("Requesting sequence initialize for '" + sequenceName + "'");
            ManagerQueue.AddMessage(new CloudQueueMessage(CreateRequestForInitializeSequence(sequenceName)));
            Trace.TraceInformation("Requested sequence initialize for '" + sequenceName + "'");

            int attemptCount = 0;

            bool queueExists = queue.Exists();

            while (queueExists == false && attemptCount < MAX_ATTEMPT_COUNT)
            {
                attemptCount++;
                Thread.Sleep(RETRY_WAIT_TIME_IN_MILLISECONDS);
                queueExists = queue.Exists();
            }

            if (queueExists == false)
            {
                throw new InvalidOperationException(String.Format("Sequence queue '{0}' failed to initialize.", sequenceName));
            }
        }

        private string CreateRequestForInitializeSequence(string sequenceName)
        {
            return String.Format("InitializeSequence:{0}", sequenceName);
        }

        public int NextValue()
        {
            CloudQueueMessage nextValueMessage = SequenceQueue.GetMessage();

            if (nextValueMessage == null)
            {
                Thread.Sleep(500);
                nextValueMessage = SequenceQueue.GetMessage();

                if (nextValueMessage == null)
                {
                    throw new InvalidOperationException("Sequence queue is empty.");
                }
            }

            SequenceQueue.DeleteMessage(nextValueMessage);

            if (nextValueMessage == null)
            {
                throw new InvalidOperationException(String.Format("Next value for sequence '{0}' was null.", Name));
            }
            else
            {
                var valueAsString = nextValueMessage.AsString;

                int asInt32 = 0;

                if (Int32.TryParse(valueAsString, out asInt32) == false)
                {
                    throw new InvalidOperationException(String.Format("Next value for sequence '{0}' was not an integer.  The value was '{1}'.", Name, valueAsString));
                }
                else
                {
                    return asInt32;
                }
            }
        }
    }
}
