using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using System.Threading;
using System.Diagnostics;

namespace Benday.SequenceGenerator.Business
{
    public class SequenceQueueManager
    {
        private const string QUEUE_NAME_SEQUENCE_MANAGER = "sequencemanager";
        private const int MINIMUM_QUEUE_DEPTH = 10;
        private const int QUEUE_REFILL_AMOUNT = 5;

        public SequenceQueueManager()
        {
            InitializeAllSequences();
        }

        private CloudQueueClient m_client;
        public CloudQueueClient Client
        {
            get
            {
                if (m_client == null)
                {
                    m_client = CloudStorageAccount.DevelopmentStorageAccount.CreateCloudQueueClient();
                }

                return m_client;
            }
        }

        private IDictionary<string, CloudQueue> m_sequenceQueues;
        public IDictionary<string, CloudQueue> SequenceQueues
        {
            get
            {
                if (m_sequenceQueues == null)
                {
                    m_sequenceQueues = new Dictionary<string, CloudQueue>();
                }

                return m_sequenceQueues;
            }
        }

        private CloudQueue m_managerQueue;
        public CloudQueue ManagerQueue
        {
            get
            {
                if (m_managerQueue == null)
                {
                    m_managerQueue = Client.GetQueueReference(QUEUE_NAME_SEQUENCE_MANAGER);
                    m_managerQueue.CreateIfNotExist();
                }

                return m_managerQueue;
            }
        }

        public void ProcessMessages()
        {
            var message = ManagerQueue.GetMessage();

            while (message != null)
            {
                ProcessMessage(message);

                message = ManagerQueue.GetMessage();
            }
        }

        private void ProcessMessage(CloudQueueMessage message)
        {
            Trace.TraceInformation(String.Format("ProcessMessage() -- '{0}'", message.AsString));

            string[] tokens = message.AsString.Split(':');

            ManagerQueue.DeleteMessage(message);

            if (tokens.Length < 2)
            {
                Trace.TraceError(String.Format("Failed to tokenize message ({0}).", message.AsString));
            }
            else
            {
                if (tokens[0] == "InitializeSequence")
                {
                    InitializeSequence(tokens[1]);
                }
                else
                {
                    Trace.TraceError(String.Format("ProcessMessage() -- Unknown message type '{0}'", tokens[0]));
                }
            }
        }

        private void InitializeSequence(string sequenceName)
        {
            Trace.TraceInformation(String.Format("InitializeSequence('{0}')", sequenceName));

            if (SequenceQueues.ContainsKey(sequenceName) == true)
            {
                Trace.TraceInformation(String.Format("InitializeSequence('{0}') -- sequence already in dictionary", sequenceName));
                CheckSequenceLength(sequenceName);
            }
            else
            {
                Trace.TraceInformation(String.Format("InitializeSequence('{0}') -- sequence not in dictionary", sequenceName));
                SequenceInformationService service = new SequenceInformationService();

                var info = service.Get(sequenceName);

                if (info == null)
                {
                    info = new SequenceInformation();
                    info.Name = sequenceName;

                    service.Save(info);
                }

                var queue = Client.GetQueueReference("sequence-" + sequenceName.ToLower());

                SequenceQueues.Add(sequenceName, queue);

                CheckSequenceLength(service, info, queue);
            }
        }

        private void CheckSequenceLength(SequenceInformationService service, SequenceInformation info, CloudQueue queue)
        {
            Trace.TraceInformation(String.Format("Checking sequence length for '{0}'.", queue.Name));

            queue.CreateIfNotExist();

            var depth = queue.RetrieveApproximateMessageCount();

            if (depth < 100)
            {
                Trace.TraceInformation(
                        "Refilling queue");

                int nextValue = info.HighestValue;

                for (int i = 0; i < QUEUE_REFILL_AMOUNT; i++)
                {
                    queue.AddMessage(new CloudQueueMessage((++nextValue).ToString()));
                }

                info.HighestValue = nextValue;

                service.Save(info);
            }
        }

        public void CheckSequenceLengths()
        {
            foreach (var sequenceName in SequenceQueues.Keys)
            {
                CheckSequenceLength(sequenceName);
            }
        }

        private void CheckSequenceLength(string sequenceName)
        {
            Trace.TraceInformation(
                        String.Format(
                        "CheckSequenceLength('{0}')",
                        sequenceName));

            if (SequenceQueues.ContainsKey(sequenceName) == false)
            {
                InitializeSequence(sequenceName);
            }

            var queue = SequenceQueues[sequenceName];

            int queueApproximateMessageCount = queue.RetrieveApproximateMessageCount();

            if (queueApproximateMessageCount < MINIMUM_QUEUE_DEPTH)
            {
                SequenceInformationService service = new SequenceInformationService();

                var info = service.Get(sequenceName);

                if (info == null)
                {
                    Trace.TraceError(
                        String.Format(
                        "CheckSequenceLength('{0}') wanted to refill the sequence queue but could not find the SequenceInformation.",
                        sequenceName));

                    return;
                }
                else
                {
                    CheckSequenceLength(service, info, queue);
                }
            }
        }

        private void InitializeAllSequences()
        {
            SequenceInformationService service = new SequenceInformationService();

            var allSequences = service.GetList();

            if (allSequences != null)
            {
                foreach (var seq in allSequences)
                {
                    InitializeSequence(seq.Name);
                }
            }
        }
    }
}
